with Text_IO; use Text_IO;
procedure s1_agoraserverada3 is

-- Claude Kaiser october 2007
-- Non-deterministic pairing 
-- followed by a parameterless call to a shared protected procedure
-- case study for Int. Conf. RST-ADA EU 08 and for QUASAR TUTORIAL
-- Shared agora algorithm 
-- usual emulation of a moniteur using an Ada protected object
-- N is the number of cooperating processes 

-- This  N = 3

   Nb_Process : constant Integer := 3;
   type Id is range 1 .. Nb_Process;
   Nb_Trial : constant Integer := 6;

   --==========================================================================
   --  concurrency types 
   --==========================================================================

   -- protected type declarations with visible entry Rendezvous and private entry Resume
  protected type T_Agora is
    entry Rendezvous(X: in Id; X_Other: out Id; Group_Leader: out Id);
  private
    entry Resume(X: in Id; X_Other: out Id; Group_Leader: out Id);
    First_Call : Boolean := True;      -- data shared by calling threads
    First_Partner : Id;
    Second_Partner : Id; 
  end T_Agora;
 
    -- pairing and peering processes type declarations
    task type T_Process is
        entry Get_Id (Y : in Id);
        -- entry Start_Peering (Pilot : in Id; Copilot : in Id; Leader : in Id);
        -- entry Finish_Peering (Copilot : in Id; Pilot : in Id; Leader : in Id);
     end T_Process;

   --===
   --  concurrent objects
   --===
    protected type T_Register is
       procedure Update;
  private 
    Count : Integer := 0;
  end T_Register;

   protected body T_Register is
    procedure Update is begin Count := Count + 1; end Update;
  end T_Register;
  
  System_Count : T_Register;
   ------------
   type Process_Array is array (Id) of T_Process;

   Process : Process_Array;
   
   Agora : T_Agora;

   --==========================================================================
   --  concurrency types body
   --==========================================================================

   protected body T_Agora is
  
    entry Rendezvous(X: in Id; X_Other: out Id; Group_Leader: out Id) when True is
      begin
        if First_Call then
          First_Partner := X; 
          First_Call := False;
          requeue Resume;
        else 
          Second_Partner := X; X_Other := First_Partner; Group_Leader := First_Partner;
          First_Call := True;
        end if;
    end Rendezvous;
    
    entry Resume(X: in Id; X_Other: out Id; Group_Leader: out Id) when First_Call is
      begin 
        X_Other := Second_Partner; 
        Group_Leader := X; -- the first calling thread is chosen as group leader
      end Resume;
  end T_Agora; 


   task body T_Process is
      Ego     : Id;
      Partner : Id;
      Leader  : Id;
      Done    : Boolean := False;
   begin
      accept Get_Id (Y : in Id) do
         Ego := Y;
      end Get_Id;
      --for I in  1 .. Nb_Trial loop
      loop
         -- Server.Cooperate (Ego, Partner, Leader);
         Agora.Rendezvous(Ego, Partner, Leader); -- seeks a partner
         Put_Line
           (" Process" &   Id'Image (Ego) &
            " is notified that its booked partner is process " &
            Id'Image (Partner) &
            "  The Group Leadership owns to " & Id'Image (Leader));
      
     Put_Line("PROCESS" & Id'Image (Ego) & " finishes PEERING with" & Id'Image (Partner));
         System_Count.Update;         
         
      end loop;
   end T_Process;

begin
   for I in  Id loop
      Process (I).Get_Id (I);
   end loop;
end s1_agoraserverada3;
