with Text_IO; use Text_IO;
procedure dummy3 is

-- Claude Kaiser october 2007
-- deterministic pairing followed by a peer to peer interaction (with 2 rendezvous)
-- case study for Int. Conf. RST-ADA EU 08 and for QUASAR TUTORIAL
-- dummy component with a shared pairing procedure providing a low bound case
-- here the partner choice is predefined and is always the same
-- Quasar provides a mean for signalling that a statement should not be sliced
-- this is done by an annotation appearing as a special comment: " --#Y"
-- N is the number of cooperating processes 

-- This  N = 3

-- Quasar current constraints
-- package names and file names in low case 
-- no embedded package in a compilation unit
-- only one declaration at a time; no identifier list (for with and use also)
-- no implicit type for task, protected object or array 
-- no explicit cast such as Integer(I)
-- when using "requeue", when possible, always assign a value to out parameter in the last step 
-- if out parameter assigned before requeueing use rather an "in out" parameter


-- Quasar residual delicacy constraints!!
-- "if (First_Call) then" causes error; "if First_Call then"  works correctly!!
-- "Max : constant := 6;" causes error; Max : constant Integer := 6; is correct!!
-- Score : Integer; case Score(C) is
--            when 0 =>    .................;
--            when 1 => ....................;
--            when 2 =>  null;
--                   end case;
-- causes error since when 2 =>  null; is sliced
-- correct solution (and also better Ada programming) 
-- case Score(C) is
--            when 0 =>    ................
--            when 1 => ....................
--            when others =>  null;
--         end case;

   Nb_Process : constant Integer := 3;
   type Id is range 1 .. Nb_Process;
   Max : Id := 3;
 
   --==========================================================================
   --  dummy rendezvous procedure 
   --==========================================================================
   procedure Rendezvous(X: in Id; X_Other: out Id; Group_Leader: out Id) --#Y
   is 
   begin
     if X /= Max then 
          X_Other := Max; Group_Leader := X ; 
     else 
        X_Other := 1 ; Group_Leader := 1;
     end if;
   end Rendezvous;
   --===
   --  concurrent objects
   --===
    task type T_Process is
        entry Get_Id (Y : in Id);
        entry Start_Peering (Pilot : in Id; Copilot : in Id; Leader : in Id);
        entry Finish_Peering (Copilot : in Id; Pilot : in Id; Leader : in Id);    
     end T_Process;
   type Process_Array is array (Id) of T_Process;

   Process : Process_Array;
   
   --==========================================================================
   --  concurrency types body
   --==========================================================================
   task body T_Process is
      Ego     : Id;
      Partner : Id;
      Leader  : Id;
   begin
      accept Get_Id (Y : in Id) do
         Ego := Y;
      end Get_Id;
    loop
         -- Server.Cooperate (Ego, Partner, Leader);
         Rendezvous(Ego, Partner, Leader); --#Y
     
          Put_Line(" Process" & Id'Image (Ego) & " is notified that its booked partner is process " 
     & Id'Image (Partner) & "  The Group Leadership owns to " & Id'Image (Leader));

         if Ego = Leader then
       -- sends intializing RPC
            Process (Partner).Start_Peering (Ego, Partner, Leader); 
     
      delay (0.01);
        -- sends finishing RPC
            Process (Partner).Finish_Peering (Ego, Partner, Leader);
 
         else
            accept Start_Peering (
               Pilot    : in Id;
               Copilot : in Id;
               Leader  : in Id)
      do 
        Put_Line
              ("PROCESS" & Id'Image (Ego) & " acts as COPILOT while PEERING with" &
               Id'Image (Pilot) &" as PILOT");
      end Start_Peering;
         delay (0.01);
       accept Finish_Peering (
              Copilot  : in Id;
               Pilot   : in Id;
               Leader  : in Id);
         end if;
         

     Put_Line("PROCESS" & Id'Image (Ego) & " finishes PEERING with" & Id'Image (Partner));
         
         
      end loop;
   end T_Process;

begin
   for I in  Id loop
      Process (I).Get_Id (I);
   end loop;
end dummy3;
