with Text_IO; use Text_IO;
procedure t1_dummy3 is

-- Claude Kaiser october 2007
-- deterministic pairing 
-- followed by a peer to peer interaction (with 2 rendezvous)
-- plus a parameterless call to a shared protected procedure
-- case study for Int. Conf. RST-ADA EU 08 and for QUASAR TUTORIAL
-- dummy component with a shared pairing procedure providing a low bound case
-- here the partner choice is predefined and is always the same
-- Quasar provides a mean for signalling that a statement should not be sliced
-- this is done by an annotation appearing as a special comment: " --#Y"
-- N is the number of cooperating processes 

-- This  N = 3

   Nb_Process : constant Integer := 3;
   type Id is range 1 .. Nb_Process;
   Max : Id := 3;
 
   --==========================================================================
   --  dummy rendezvous procedure 
   --==========================================================================
   procedure Rendezvous(X: in Id; X_Other: out Id; Group_Leader: out Id) --#Y
   is 
   begin
     if X /= Max then 
          X_Other := Max; Group_Leader := X ; 
     else 
        X_Other := 1 ; Group_Leader := 1;
     end if;
   end Rendezvous;
   --===
   --  concurrent objects
   --===
    protected type T_Register is
       procedure Update;
  private 
    Count : Integer := 0;
  end T_Register;

   protected body T_Register is
    procedure Update is begin Count := Count + 1; end Update;
  end T_Register;
  
  System_Count : T_Register;
  
    task type T_Process is
        entry Get_Id (Y : in Id);
        entry Start_Peering (Pilot : in Id; Copilot : in Id; Leader : in Id);
    entry Finish_Peering (Copilot : in Id; Pilot : in Id; Leader : in Id);    
     end T_Process;
   type Process_Array is array (Id) of T_Process;

   Process : Process_Array;
   
   --==========================================================================
   --  concurrency types body
   --==========================================================================
   task body T_Process is
      Ego     : Id;
      Partner : Id;
      Leader  : Id;
   begin
      accept Get_Id (Y : in Id) do
         Ego := Y;
      end Get_Id;
    loop
         -- Server.Cooperate (Ego, Partner, Leader);
         Rendezvous(Ego, Partner, Leader); --#Y
     
          Put_Line(" Process" & Id'Image (Ego) & " is notified that its booked partner is process " 
     & Id'Image (Partner) & "  The Group Leadership owns to " & Id'Image (Leader));

         if Ego = Leader then
       -- sends intializing RPC
            Process (Partner).Start_Peering (Ego, Partner, Leader); 
     
      delay (0.01);
        -- sends finishing RPC
             Process (Partner).Finish_Peering (Ego, Partner, Leader);
 
         else
            accept Start_Peering (
               Pilot    : in Id;
               Copilot : in Id;
               Leader  : in Id)
      do 
        Put_Line
              ("PROCESS" & Id'Image (Ego) & " acts as COPILOT while PEERING with" &
               Id'Image (Pilot) &" as PILOT");
      end Start_Peering;
         delay (0.01);
        accept Finish_Peering (
               Copilot  : in Id;
               Pilot   : in Id;
               Leader  : in Id);
         end if;
         

     Put_Line("PROCESS" & Id'Image (Ego) & " finishes PEERING with" & Id'Image (Partner));
         
     System_Count.Update; 
         
      end loop;
   end T_Process;

begin
   for I in  Id loop
      Process (I).Get_Id (I);
   end loop;
end t1_dummy3;
